# Risk.API.Client - the C# library for the Risk.API

Risk Web API

This C# SDK is automatically generated by the [OpenAPI Generator](https://openapi-generator.tech) project:

- API version: v0.1.0
- SDK version: 1.0
- Build package: org.openapitools.codegen.languages.CSharpNetCoreClientCodegen
    For more information, please visit [https://github.com/jtsoya539](https://github.com/jtsoya539)

<a id="frameworks-supported"></a>
## Frameworks supported
- .NET Core >=1.0
- .NET Framework >=4.6
- Mono/Xamarin >=vNext

<a id="dependencies"></a>
## Dependencies

- [RestSharp](https://www.nuget.org/packages/RestSharp) - 106.13.0 or later
- [Json.NET](https://www.nuget.org/packages/Newtonsoft.Json/) - 13.0.2 or later
- [JsonSubTypes](https://www.nuget.org/packages/JsonSubTypes/) - 1.8.0 or later
- [System.ComponentModel.Annotations](https://www.nuget.org/packages/System.ComponentModel.Annotations) - 5.0.0 or later

The DLLs included in the package may not be the latest version. We recommend using [NuGet](https://docs.nuget.org/consume/installing-nuget) to obtain the latest version of the packages:
```
Install-Package RestSharp
Install-Package Newtonsoft.Json
Install-Package JsonSubTypes
Install-Package System.ComponentModel.Annotations
```

NOTE: RestSharp versions greater than 105.1.0 have a bug which causes file uploads to fail. See [RestSharp#742](https://github.com/restsharp/RestSharp/issues/742).
NOTE: RestSharp for .Net Core creates a new socket for each api call, which can lead to a socket exhaustion problem. See [RestSharp#1406](https://github.com/restsharp/RestSharp/issues/1406).

<a id="installation"></a>
## Installation
Generate the DLL using your preferred tool (e.g. `dotnet build`)

Then include the DLL (under the `bin` folder) in the C# project, and use the namespaces:
```csharp
using Risk.API.Client.Api;
using Risk.API.Client.Client;
using Risk.API.Client.Model;
```
<a id="usage"></a>
## Usage

To use the API client with a HTTP proxy, setup a `System.Net.WebProxy`
```csharp
Configuration c = new Configuration();
System.Net.WebProxy webProxy = new System.Net.WebProxy("http://myProxyUrl:80/");
webProxy.Credentials = System.Net.CredentialCache.DefaultCredentials;
c.Proxy = webProxy;
```

<a id="getting-started"></a>
## Getting Started

```csharp
using System.Collections.Generic;
using System.Diagnostics;
using Risk.API.Client.Api;
using Risk.API.Client.Client;
using Risk.API.Client.Model;

namespace Example
{
    public class Example
    {
        public static void Main()
        {

            Configuration config = new Configuration();
            config.BasePath = "https://localhost:5001";
            var apiInstance = new AutApi(config);
            var key = "key_example";  // string | Clave para la activación
            var riskDeviceToken = "riskDeviceToken_example";  // string | Token del dispositivo desde el cual se realiza la petición (optional) 
            var riskServiceVersion = "riskServiceVersion_example";  // string | Versión del servicio (optional) 

            try
            {
                // ActivarUsuario
                DatoRespuesta result = apiInstance.ActivarUsuario(key, riskDeviceToken, riskServiceVersion);
                Debug.WriteLine(result);
            }
            catch (ApiException e)
            {
                Debug.Print("Exception when calling AutApi.ActivarUsuario: " + e.Message );
                Debug.Print("Status Code: "+ e.ErrorCode);
                Debug.Print(e.StackTrace);
            }

        }
    }
}
```

<a id="documentation-for-api-endpoints"></a>
## Documentation for API Endpoints

All URIs are relative to *https://localhost:5001*

Class | Method | HTTP request | Description
------------ | ------------- | ------------- | -------------
*AutApi* | [**ActivarUsuario**](docs\AutApi.md#activarusuario) | **GET** /Aut/ActivarUsuario | ActivarUsuario
*AutApi* | [**CambiarClaveAcceso**](docs\AutApi.md#cambiarclaveacceso) | **POST** /Api/Aut/CambiarClaveAcceso | CambiarClaveAcceso
*AutApi* | [**CambiarClaveTransaccional**](docs\AutApi.md#cambiarclavetransaccional) | **POST** /Api/Aut/CambiarClaveTransaccional | CambiarClaveTransaccional
*AutApi* | [**DatosUsuario**](docs\AutApi.md#datosusuario) | **GET** /Api/Aut/DatosUsuario | DatosUsuario
*AutApi* | [**EditarDatoUsuario**](docs\AutApi.md#editardatousuario) | **POST** /Api/Aut/EditarDatoUsuario | EditarDatoUsuario
*AutApi* | [**EditarUsuario**](docs\AutApi.md#editarusuario) | **POST** /Api/Aut/EditarUsuario | EditarUsuario
*AutApi* | [**EliminarUsuario**](docs\AutApi.md#eliminarusuario) | **POST** /Api/Aut/EliminarUsuario | EliminarUsuario
*AutApi* | [**FinalizarSesion**](docs\AutApi.md#finalizarsesion) | **POST** /Api/Aut/FinalizarSesion | FinalizarSesion
*AutApi* | [**GenerarOtp**](docs\AutApi.md#generarotp) | **POST** /Api/Aut/GenerarOtp | GenerarOtp
*AutApi* | [**GuardarAvatarUsuario**](docs\AutApi.md#guardaravatarusuario) | **POST** /Api/Aut/GuardarAvatarUsuario | GuardarAvatarUsuario
*AutApi* | [**IniciarSesion**](docs\AutApi.md#iniciarsesion) | **POST** /Api/Aut/IniciarSesion | IniciarSesion
*AutApi* | [**IniciarSesionFacebook**](docs\AutApi.md#iniciarsesionfacebook) | **POST** /Api/Aut/IniciarSesionFacebook | IniciarSesionFacebook
*AutApi* | [**IniciarSesionGoogle**](docs\AutApi.md#iniciarsesiongoogle) | **POST** /Api/Aut/IniciarSesionGoogle | IniciarSesionGoogle
*AutApi* | [**RecuperarAvatarUsuario**](docs\AutApi.md#recuperaravatarusuario) | **GET** /Api/Aut/RecuperarAvatarUsuario | RecuperarAvatarUsuario
*AutApi* | [**RefrescarSesion**](docs\AutApi.md#refrescarsesion) | **POST** /Api/Aut/RefrescarSesion | RefrescarSesion
*AutApi* | [**RefrescarSesionFacebook**](docs\AutApi.md#refrescarsesionfacebook) | **POST** /Api/Aut/RefrescarSesionFacebook | RefrescarSesionFacebook
*AutApi* | [**RefrescarSesionGoogle**](docs\AutApi.md#refrescarsesiongoogle) | **POST** /Api/Aut/RefrescarSesionGoogle | RefrescarSesionGoogle
*AutApi* | [**RegistrarClaveTransaccional**](docs\AutApi.md#registrarclavetransaccional) | **POST** /Api/Aut/RegistrarClaveTransaccional | RegistrarClaveTransaccional
*AutApi* | [**RegistrarDispositivo**](docs\AutApi.md#registrardispositivo) | **POST** /Api/Aut/RegistrarDispositivo | RegistrarDispositivo
*AutApi* | [**RegistrarUbicacion**](docs\AutApi.md#registrarubicacion) | **POST** /Api/Aut/RegistrarUbicacion | RegistrarUbicacion
*AutApi* | [**RegistrarUsuario**](docs\AutApi.md#registrarusuario) | **POST** /Api/Aut/RegistrarUsuario | RegistrarUsuario
*AutApi* | [**ValidarOtp**](docs\AutApi.md#validarotp) | **GET** /Api/Aut/ValidarOtp | ValidarOtp
*AutApi* | [**ValidarPermiso**](docs\AutApi.md#validarpermiso) | **GET** /Api/Aut/ValidarPermiso | ValidarPermiso
*AutApi* | [**ValidarSesion**](docs\AutApi.md#validarsesion) | **GET** /Api/Aut/ValidarSesion | ValidarSesion
*GenApi* | [**GuardarArchivo**](docs\GenApi.md#guardararchivo) | **POST** /Api/Gen/GuardarArchivo | GuardarArchivo
*GenApi* | [**ListarAplicaciones**](docs\GenApi.md#listaraplicaciones) | **GET** /Api/Gen/ListarAplicaciones | ListarAplicaciones
*GenApi* | [**ListarErrores**](docs\GenApi.md#listarerrores) | **GET** /Api/Gen/ListarErrores | ListarErrores
*GenApi* | [**ListarSignificados**](docs\GenApi.md#listarsignificados) | **GET** /Api/Gen/ListarSignificados | ListarSignificados
*GenApi* | [**RecuperarArchivo**](docs\GenApi.md#recuperararchivo) | **GET** /Api/Gen/RecuperarArchivo | RecuperarArchivo
*GenApi* | [**RecuperarTexto**](docs\GenApi.md#recuperartexto) | **GET** /Api/Gen/RecuperarTexto | RecuperarTexto
*GenApi* | [**ReporteListarSignificados**](docs\GenApi.md#reportelistarsignificados) | **GET** /Api/Gen/ReporteListarSignificados | ReporteListarSignificados
*GenApi* | [**ReporteVersionSistema**](docs\GenApi.md#reporteversionsistema) | **GET** /Api/Gen/ReporteVersionSistema | ReporteVersionSistema
*GenApi* | [**SignificadoCodigo**](docs\GenApi.md#significadocodigo) | **GET** /Api/Gen/SignificadoCodigo | SignificadoCodigo
*GenApi* | [**ValorParametro**](docs\GenApi.md#valorparametro) | **GET** /Api/Gen/ValorParametro | ValorParametro
*GenApi* | [**VersionServicio**](docs\GenApi.md#versionservicio) | **GET** /Api/Gen/VersionServicio | VersionServicio
*GenApi* | [**VersionSistema**](docs\GenApi.md#versionsistema) | **GET** /Gen/VersionSistema | VersionSistema
*GloApi* | [**ListarBarrios**](docs\GloApi.md#listarbarrios) | **GET** /Api/Glo/ListarBarrios | ListarBarrios
*GloApi* | [**ListarCiudades**](docs\GloApi.md#listarciudades) | **GET** /Api/Glo/ListarCiudades | ListarCiudades
*GloApi* | [**ListarDepartamentos**](docs\GloApi.md#listardepartamentos) | **GET** /Api/Glo/ListarDepartamentos | ListarDepartamentos
*GloApi* | [**ListarPaises**](docs\GloApi.md#listarpaises) | **GET** /Api/Glo/ListarPaises | ListarPaises
*MsjApi* | [**ActivarMensajeria**](docs\MsjApi.md#activarmensajeria) | **POST** /Api/Msj/ActivarMensajeria | ActivarMensajeria
*MsjApi* | [**CambiarEstadoMensajeria**](docs\MsjApi.md#cambiarestadomensajeria) | **POST** /Api/Msj/CambiarEstadoMensajeria | CambiarEstadoMensajeria
*MsjApi* | [**DesactivarMensajeria**](docs\MsjApi.md#desactivarmensajeria) | **POST** /Api/Msj/DesactivarMensajeria | DesactivarMensajeria
*MsjApi* | [**ListarCorreosPendientes**](docs\MsjApi.md#listarcorreospendientes) | **GET** /Api/Msj/ListarCorreosPendientes | ListarCorreosPendientes
*MsjApi* | [**ListarMensajesPendientes**](docs\MsjApi.md#listarmensajespendientes) | **GET** /Api/Msj/ListarMensajesPendientes | ListarMensajesPendientes
*MsjApi* | [**ListarNotificacionesPendientes**](docs\MsjApi.md#listarnotificacionespendientes) | **GET** /Api/Msj/ListarNotificacionesPendientes | ListarNotificacionesPendientes
*RepApi* | [**ReporteListarSignificados**](docs\RepApi.md#reportelistarsignificados) | **GET** /Api/Gen/ReporteListarSignificados | ReporteListarSignificados
*RepApi* | [**ReporteVersionSistema**](docs\RepApi.md#reporteversionsistema) | **GET** /Api/Gen/ReporteVersionSistema | ReporteVersionSistema


<a id="documentation-for-models"></a>
## Documentation for Models

 - [Model.AccionPermiso](docs\AccionPermiso.md)
 - [Model.Aplicacion](docs\Aplicacion.md)
 - [Model.AplicacionPagina](docs\AplicacionPagina.md)
 - [Model.AplicacionPaginaRespuesta](docs\AplicacionPaginaRespuesta.md)
 - [Model.Archivo](docs\Archivo.md)
 - [Model.Barrio](docs\Barrio.md)
 - [Model.BarrioPagina](docs\BarrioPagina.md)
 - [Model.BarrioPaginaRespuesta](docs\BarrioPaginaRespuesta.md)
 - [Model.CambiarClaveAccesoRequestBody](docs\CambiarClaveAccesoRequestBody.md)
 - [Model.CambiarClaveTransaccionalRequestBody](docs\CambiarClaveTransaccionalRequestBody.md)
 - [Model.CambiarEstadoMensajeriaRequestBody](docs\CambiarEstadoMensajeriaRequestBody.md)
 - [Model.Ciudad](docs\Ciudad.md)
 - [Model.CiudadPagina](docs\CiudadPagina.md)
 - [Model.CiudadPaginaRespuesta](docs\CiudadPaginaRespuesta.md)
 - [Model.Correo](docs\Correo.md)
 - [Model.CorreoPagina](docs\CorreoPagina.md)
 - [Model.CorreoPaginaRespuesta](docs\CorreoPaginaRespuesta.md)
 - [Model.Dato](docs\Dato.md)
 - [Model.DatoRespuesta](docs\DatoRespuesta.md)
 - [Model.Departamento](docs\Departamento.md)
 - [Model.DepartamentoPagina](docs\DepartamentoPagina.md)
 - [Model.DepartamentoPaginaRespuesta](docs\DepartamentoPaginaRespuesta.md)
 - [Model.Dispositivo](docs\Dispositivo.md)
 - [Model.EditarDatoUsuarioRequestBody](docs\EditarDatoUsuarioRequestBody.md)
 - [Model.EditarUsuarioRequestBody](docs\EditarUsuarioRequestBody.md)
 - [Model.EliminarUsuarioRequestBody](docs\EliminarUsuarioRequestBody.md)
 - [Model.Error](docs\Error.md)
 - [Model.ErrorPagina](docs\ErrorPagina.md)
 - [Model.ErrorPaginaRespuesta](docs\ErrorPaginaRespuesta.md)
 - [Model.EstadoMensajeria](docs\EstadoMensajeria.md)
 - [Model.EstadoSesion](docs\EstadoSesion.md)
 - [Model.EstadoUsuario](docs\EstadoUsuario.md)
 - [Model.FinalizarSesionRequestBody](docs\FinalizarSesionRequestBody.md)
 - [Model.FormatoReporte](docs\FormatoReporte.md)
 - [Model.IniciarSesionFacebookRequestBody](docs\IniciarSesionFacebookRequestBody.md)
 - [Model.IniciarSesionGoogleRequestBody](docs\IniciarSesionGoogleRequestBody.md)
 - [Model.IniciarSesionRequestBody](docs\IniciarSesionRequestBody.md)
 - [Model.Mensaje](docs\Mensaje.md)
 - [Model.MensajePagina](docs\MensajePagina.md)
 - [Model.MensajePaginaRespuesta](docs\MensajePaginaRespuesta.md)
 - [Model.Notificacion](docs\Notificacion.md)
 - [Model.NotificacionPagina](docs\NotificacionPagina.md)
 - [Model.NotificacionPaginaRespuesta](docs\NotificacionPaginaRespuesta.md)
 - [Model.OrigenSesion](docs\OrigenSesion.md)
 - [Model.Pais](docs\Pais.md)
 - [Model.PaisPagina](docs\PaisPagina.md)
 - [Model.PaisPaginaRespuesta](docs\PaisPaginaRespuesta.md)
 - [Model.Plantilla](docs\Plantilla.md)
 - [Model.RefrescarSesionFacebookRequestBody](docs\RefrescarSesionFacebookRequestBody.md)
 - [Model.RefrescarSesionGoogleRequestBody](docs\RefrescarSesionGoogleRequestBody.md)
 - [Model.RefrescarSesionRequestBody](docs\RefrescarSesionRequestBody.md)
 - [Model.RegistrarClaveTransaccionalRequestBody](docs\RegistrarClaveTransaccionalRequestBody.md)
 - [Model.RegistrarDispositivoRequestBody](docs\RegistrarDispositivoRequestBody.md)
 - [Model.RegistrarUbicacionRequestBody](docs\RegistrarUbicacionRequestBody.md)
 - [Model.RegistrarUsuarioRequestBody](docs\RegistrarUsuarioRequestBody.md)
 - [Model.Rol](docs\Rol.md)
 - [Model.Sesion](docs\Sesion.md)
 - [Model.SesionRespuesta](docs\SesionRespuesta.md)
 - [Model.Significado](docs\Significado.md)
 - [Model.SignificadoPagina](docs\SignificadoPagina.md)
 - [Model.SignificadoPaginaRespuesta](docs\SignificadoPaginaRespuesta.md)
 - [Model.TipoAplicacion](docs\TipoAplicacion.md)
 - [Model.TipoDispositivo](docs\TipoDispositivo.md)
 - [Model.TipoMensajeria](docs\TipoMensajeria.md)
 - [Model.Usuario](docs\Usuario.md)
 - [Model.UsuarioRespuesta](docs\UsuarioRespuesta.md)


<a id="documentation-for-authorization"></a>
## Documentation for Authorization


Authentication schemes defined for the API:
<a id="RiskAppKey"></a>
### RiskAppKey

- **Type**: API key
- **API key parameter name**: Risk-App-Key
- **Location**: HTTP header

<a id="AccessToken"></a>
### AccessToken

- **Type**: Bearer Authentication

